/*
 * QEstEidBreakpad
 *
 * Copyright (C) 2012 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2012 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

/*
 * dump_syms qbreakpad > qbreakpad.sym
 * head -n1 qbreakpad.sym
 * "MODULE mac x86_64 1F902FC241353FE28AA37FAEDB23131E0 qbreakpad"
 * mkdir -p sympath/qbreakpad/1F902FC241353FE28AA37FAEDB23131E0/
 * mv qbreakpad.sym sympath/qbreakpad/1F902FC241353FE28AA37FAEDB23131E0/
 *
 * Debug
 * minidump_stackwalk 219F15E5-1B1F-4F08-84D8-28031706FD81.dmp ./symbols/ 2>&1
 */

#include "QBreakPad.h"

#include <common/Settings.h>

#include <QCoreApplication>
#include <QDebug>
#include <QDir>
#include <QFile>
#include <QHttpMultiPart>
#include <QLabel>
#include <QMessageBox>
#include <QNetworkAccessManager>
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QProcess>
#include <QPlainTextEdit>
#include <QTranslator>
#include <QUrl>
#include <QVBoxLayout>

#if defined(Q_OS_MAC)
#include "client/mac/handler/exception_handler.h"
#elif defined(Q_OS_LINUX)
#include "client/linux/handler/exception_handler.h"
#elif defined(Q_OS_WIN)
#include "client/windows/handler/exception_handler.h"
#endif

#ifdef Q_OS_WIN
static bool QBreakPadDump( const wchar_t *dump_path, const wchar_t *minidump_id, void *,
	EXCEPTION_POINTERS *, MDRawAssertionInfo *, bool succeeded )
{
	qDebug() << "Application crashed, starting crashreport application";
	QProcess::startDetached( qApp->applicationFilePath(), QStringList() << "-crashreport"
		<< QString::fromUtf16( (const ushort*)dump_path )
		<< QString::fromUtf16( (const ushort*)minidump_id ) );
	return succeeded;
}
#else
static bool QBreakPadDump( const char *dump_path, const char *minidump_id, void *, bool succeeded )
{
	qDebug() << "Application crashed, starting crashreport application";
	QProcess::startDetached( qApp->applicationFilePath(), QStringList() << "-crashreport"
		<< dump_path << minidump_id );
	return succeeded;
}
#endif

QBreakPad::QBreakPad()
	: d( 0 )
{
	qDebug() << "Enabling crashreporting";
#if defined(Q_OS_WIN)
	QString tmp = QDir::tempPath();
	d = new google_breakpad::ExceptionHandler( LPCWSTR(tmp.utf16()), 0, QBreakPadDump, 0,
		google_breakpad::ExceptionHandler::HANDLER_ALL );
#elif defined(Q_OS_MAC)
	QByteArray tmp = QDir::tempPath().toLocal8Bit();
	d = new google_breakpad::ExceptionHandler( tmp.constData(), 0, QBreakPadDump, 0, true, 0 );
#else
	QByteArray tmp = QDir::tempPath().toLocal8Bit();
	d = new google_breakpad::ExceptionHandler( tmp.constData(), 0, QBreakPadDump, 0, true );
#endif
}

QBreakPad::~QBreakPad()
{
	delete d;
}

bool QBreakPad::isCrashReport( int argc, char *argv[] )
{
	QString crash( "-crashreport" );
	for( int i = 0; i < argc; ++i )
		if( crash == argv[i] )
			return true;
	return false;
}

QBreakPadDialog::QBreakPadDialog( const QString &name )
	: QWizard()
{
	QTranslator *qt = new QTranslator( this );
	QTranslator *bp = new QTranslator( this );
	qt->load( ":/translations/qt_" + Settings::language() );
	bp->load( ":/translations/breakpad_" + Settings::language() );
	qApp->installTranslator( qt );
	qApp->installTranslator( bp );

	qDebug() << "Crashreport application started";
	setWindowTitle( tr("%1 has stopped working").arg( name ) );
	setButtonLayout( QList<WizardButton>()
		<< Stretch << CancelButton << FinishButton );
	setButtonText( FinishButton, tr("Send report") );
	setButtonText( CancelButton, tr("Close the program") );

	QWizardPage *crash = new QWizardPage( this );
	crash->setTitle( tr("Error Report") );
	crash->setSubTitle( tr("Please send us this error report to "
		"help fix the problem and improve this software.") );
	addPage( crash );

	QLabel *privacy = new QLabel( QString( "<a href='%1'>%2</a>" )
		.arg( tr("https://installer.id.ee/Privacy_Policy_ENG.html") )
		.arg( tr("ID software error reporting principles") ) );
	privacy->setOpenExternalLinks( true );
	QLabel *link = new QLabel( QString( "<a href='show'>%1</a>" ).arg(
		tr("Provide additional info about the problem (recommended)") ), crash );
	edit = new QPlainTextEdit( crash );
	edit->hide();
	connect( link, SIGNAL(linkActivated(QString)), SLOT(toggleComments()) );

	QVBoxLayout *l = new QVBoxLayout( crash );
	l->addWidget( privacy );
	l->addWidget( link );
	l->addWidget( edit );
}

void QBreakPadDialog::toggleComments()
{
	edit->setVisible( !edit->isVisible() );
}

bool QBreakPadDialog::validateCurrentPage()
{
	QStringList args = qApp->arguments();
	int pos = args.indexOf( "-crashreport" );
	if( pos == -1 )
		return true;

	QFile f( QString( "%1/%2.dmp" ).arg( args[pos + 1], args[pos + 2] ) );
	if( !f.open( QFile::ReadOnly ) )
		return true;

	QHttpPart textPart;
	textPart.setHeader( QNetworkRequest::ContentDispositionHeader, "form-data; name=\"comment\"" );
	textPart.setBody( edit->toPlainText().toUtf8() );

	QHttpPart dumpPart;
	dumpPart.setHeader( QNetworkRequest::ContentDispositionHeader, "form-data; name=\"dump\"; filename=\"dump\"" );
	dumpPart.setHeader( QNetworkRequest::ContentTypeHeader, "application/octet-stream" );
	dumpPart.setBodyDevice( &f );

	QHttpMultiPart multiPart( QHttpMultiPart::FormDataType );
	multiPart.append( textPart );
	multiPart.append( dumpPart );

	QNetworkRequest req( QUrl( BREAKPAD ) );
	req.setRawHeader( "User-Agent", property("User-Agent").toByteArray() );
	req.setRawHeader( "X-Dump-ID", args[pos + 2].toUtf8() );

	qDebug() << "uploading crash dump file" << f.fileName();
	QNetworkAccessManager manager;
	QEventLoop e;
	QNetworkReply *repl = manager.post( req, &multiPart );
	QObject::connect( repl, SIGNAL(finished()), &e, SLOT(quit()) );
	e.exec();
	if( repl->error() == QNetworkReply::NoError )
	{
		f.remove();
		return true;
	}
	else
	{
		QMessageBox::warning( this, tr("Error Report"), repl->errorString() );
		return false;
	}
}
