/*
 * QEstEidCommon
 *
 * Copyright (C) 2011-2012 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2011-2012 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "QPCSC.h"
#include "QPCSC_p.h"

#include <QDebug>
#include <QStringList>

QPCSCPrivate::QPCSCPrivate()
#if defined(Q_OS_WIN)
	: lib( "winscard" )
#elif defined(Q_OS_MAC)
	: lib( "PCSC.framework/PCSC" )
#else
	: lib( "pcsclite", 1 )
#endif
	, context(0)
{
	SCardEstablishContext = sym_SCardEstablishContext(lib.resolve("SCardEstablishContext"));
	if( !SCardEstablishContext )
		return;
	SCardReleaseContext = sym_SCardReleaseContext(lib.resolve("SCardReleaseContext"));
	SCardListReaders = sym_SCardListReaders(lib.resolve("SCardListReaders" SUFFIX) );
	SCardGetStatusChange = sym_SCardGetStatusChange(lib.resolve("SCardGetStatusChange" SUFFIX));
	SCardConnect = sym_SCardConnect(lib.resolve("SCardConnect" SUFFIX));
	SCardReconnect = sym_SCardReconnect(lib.resolve("SCardReconnect"));
	SCardDisconnect = sym_SCardDisconnect(lib.resolve("SCardDisconnect"));
#ifdef Q_OS_MAC
	SCardControl = sym_SCardControl(lib.resolve("SCardControl132"));
#else
	SCardControl = sym_SCardControl(lib.resolve("SCardControl"));
#endif
}



QPCSCReaderPrivate::QPCSCReaderPrivate(	QPCSCPrivate *_d )
	: d( _d )
	, card( 0 )
	, proto( 0 )
	, verify_ioctl( 0 )
	, verify_ioctl_start( 0 )
	, verify_ioctl_finish( 0 )
	, modify_ioctl( 0 )
	, modify_ioctl_start( 0 )
	, modify_ioctl_finish( 0 )
	, display( false )
{}

bool QPCSCReaderPrivate::updateState()
{
	if( !d->context )
		return false;
	LONG err = d->SCardGetStatusChange( d->context, 0, &state, 1 );
	return err == SCARD_S_SUCCESS;
}



QPCSC::QPCSC( QObject *parent )
	: QObject( parent )
	, d( new QPCSCPrivate )
{
	if( d->SCardEstablishContext )
		d->SCardEstablishContext( SCARD_SCOPE_USER, 0, 0, &d->context );
	else
		qDebug() << d->lib.errorString();
}

QPCSC::~QPCSC()
{
	qDeleteAll( findChildren<QPCSCReader*>() );
	if( d->context )
		d->SCardReleaseContext( d->context );
	delete d;
}

QStringList QPCSC::readers() const
{
	if( !d->context )
		return QStringList();

	DWORD size = 0;
	LONG result = d->SCardListReaders( d->context, 0, 0, &size );
	if( result != SCARD_S_SUCCESS || !size )
		return QStringList();

	QByteArray data( size, 0 );
	result = d->SCardListReaders( d->context, 0, data.data(), &size );
	if( result != SCARD_S_SUCCESS )
		return QStringList();

	data.resize( size );
	QStringList readers;
	for( QByteArray::const_iterator i = data.constBegin(); i != data.constEnd(); ++i )
	{
		QString reader( i );
		if( !reader.isEmpty() )
			readers << reader;
		i += reader.size();
	}

	return readers;
}



QPCSCReader::QPCSCReader( const QString &reader, QPCSC *parent )
	: QObject( parent )
	, d( new QPCSCReaderPrivate( parent->d ) )
{
	d->reader = reader.toUtf8();
	qMemSet( &d->state, 0, sizeof(d->state) );
	d->state.szReader = d->reader.constData();
	if( !d->updateState() )
		return;

	if( !d->d->SCardControl ) // PC/SC API is too old to support PinPADs
		return;

	/* Use DIRECT mode only if there is no card in the reader */
	if( !isPresent() )
	{
#ifndef Q_OS_WIN  /* Apple 10.5.7 and pcsc-lite previous to v1.5.5 do not support 0 as protocol identifier */
		LONG rv = d->d->SCardConnect( d->d->context, (CSTRTYPE)d->state.szReader, SCARD_SHARE_DIRECT, SCARD_PROTOCOL_T0|SCARD_PROTOCOL_T1, &d->card, &d->proto );
#else
		LONG rv = d->d->SCardConnect( d->d->context, (CSTRTYPE)d->state.szReader, SCARD_SHARE_DIRECT, 0, &d->card, &d->proto );
#endif
		// Assume that there is a card in the reader in shared mode if direct communcation failed
		if( rv == (LONG)SCARD_E_SHARING_VIOLATION && !connect() )
			return;
	}
	else if( !connect() )
		return;

	DWORD feature_len = 0;
	BYTE feature_buf[256];
	LONG rv = d->d->SCardControl( d->card, CM_IOCTL_GET_FEATURE_REQUEST, 0, 0, feature_buf, sizeof(feature_buf), &feature_len);
	if( rv == SCARD_S_SUCCESS && (feature_len % sizeof(PCSC_TLV_STRUCTURE)) == 0 )
	{
		feature_len /= sizeof(PCSC_TLV_STRUCTURE);
		PCSC_TLV_STRUCTURE *pcsc_tlv = (PCSC_TLV_STRUCTURE *)feature_buf;
		for ( DWORD i = 0; i < feature_len; i++)
		{
			switch( pcsc_tlv[i].tag )
			{
			case FEATURE_VERIFY_PIN_DIRECT:
				d->verify_ioctl = ntohl( pcsc_tlv[i].value ); break;
			case FEATURE_VERIFY_PIN_START:
				d->verify_ioctl_start = ntohl( pcsc_tlv[i].value ); break;
			case FEATURE_VERIFY_PIN_FINISH:
				d->verify_ioctl_finish = ntohl( pcsc_tlv[i].value ); break;
			case FEATURE_MODIFY_PIN_DIRECT:
				d->modify_ioctl = ntohl( pcsc_tlv[i].value ); break;
			case FEATURE_MODIFY_PIN_START:
				d->modify_ioctl_start = ntohl( pcsc_tlv[i].value ); break;
			case FEATURE_MODIFY_PIN_FINISH:
				d->modify_ioctl_finish = ntohl( pcsc_tlv[i].value ); break;
			case FEATURE_IFD_PIN_PROPERTIES:
			{
				DWORD rcount = 0;
				BYTE rbuf[256];
				rv = d->d->SCardControl( d->card, ntohl(pcsc_tlv[i].value), 0, 0, rbuf, sizeof(rbuf), &rcount );
				if( rv == SCARD_S_SUCCESS )
				{
					PIN_PROPERTIES_STRUCTURE *caps = (PIN_PROPERTIES_STRUCTURE *)rbuf;
					d->display = caps->wLcdLayout > 0;
				}
				break;
			}
			default: break;
			}
		}
	}
	disconnect();
}

QPCSCReader::~QPCSCReader()
{
	disconnect();
	delete d;
}

QByteArray QPCSCReader::atr() const
{
	return QByteArray( (const char*)d->state.rgbAtr, d->state.cbAtr ).toHex().toUpper();
}

bool QPCSCReader::connect()
{
	if( !d->d->context )
		return false;
	LONG err = d->d->SCardConnect( d->d->context, (CSTRTYPE)d->state.szReader, SCARD_SHARE_SHARED,
		SCARD_PROTOCOL_T0|SCARD_PROTOCOL_T1, &d->card, &d->proto );
	d->updateState();
	return err == SCARD_S_SUCCESS;
}

void QPCSCReader::disconnect( Reset reset )
{
	if( d->card )
		d->d->SCardDisconnect( d->card, reset );
	d->proto = 0;
	d->card = 0;
	d->updateState();
}

bool QPCSCReader::isPinPad() const
{
	return (d->verify_ioctl || (d->verify_ioctl_start && d->verify_ioctl_finish)) &&
		 (d->modify_ioctl || (d->modify_ioctl_start && d->modify_ioctl_finish));
}

bool QPCSCReader::isPresent() const
{
	return d->state.dwEventState & SCARD_STATE_PRESENT;
}

QString QPCSCReader::name() const
{
	return QString::fromUtf8( d->reader );
}

int QPCSCReader::protocol() const
{
	return d->proto;
}

bool QPCSCReader::reconnect( Reset reset )
{
	if( !d->card )
		return false;
	LONG err = d->d->SCardReconnect( d->card, SCARD_SHARE_SHARED,
		SCARD_PROTOCOL_T0|SCARD_PROTOCOL_T1, reset, &d->proto );
	d->updateState();
	return err == SCARD_S_SUCCESS;
}
